-- -*-coding: utf-8;-*-

--[[
-- ************************************************************************
-- BuddyName.lua
-- ************************************************************************
-- This addon is released under GPL v3. All additionally used libraries are
-- released under the license that accompanies them.
--]]

BuddyName = LibStub("AceAddon-3.0"):NewAddon("BuddyName",
                                             "AceConsole-3.0",
                                             "AceHook-3.0",
                                             "AceTimer-3.0",
                                             "AceComm-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("BuddyName");

local libSer     = LibStub:GetLibrary("AceSerializer-3.0");
local libComp    = LibStub:GetLibrary("LibCompress");
local libCompEnc = libComp:GetAddonEncodeTable();

local slashCommands = {
	 ["bn"]        = true,
	 ["buddyname"] = true,
	 ["BuddyName"] = true,
}

-- perhaps used in future versions
local registerEvents = {
	 ["GUILD_JOIN"] = false,
}

-- Constants used for synchronization (Magic Number Alert!)
local COMM = {
	 SYNC_PREFIX = "BuddyName",
	 PROTOCOL = 2,
	 UNKNOWN_PROTOCOL = -1,
	 SEP = { -- separator
			HEADER  = "!",
			ENTRY   = ";",
			FIELD   = ":",
	 },
	 SECTION_TYPE = {
			ENTRY_SET        = 1,
			HASH_PROPAGATION = 2,
			DATA_REQUEST     = 3,
			UNKNOWN          = -1,
	 },
	 COMPRESSION = {
			NONE        = 0,
			LIBCOMPRESS = 1,
	 },
	 HEADER = {
			PREFIX="BN",
	 },
	 DATA_DOMAIN = {
			REALM      = 1,
			GUILD      = 2,
			GUILD_DIFF = 3,
	 },
	 ENTRY_SET_TYPE = {
			FULL_DUMP    = 1,
			DIFF_DUMP    = 2,
			SINGLE_ENTRY = 3,
	 },
	 RESPONSE_METHOD = {
			GUILD   = 1,
			WHISPER = 2,
	 },
	 IS_NO_RESPONSE = 0,
	 IS_RESPONSE    = 1,
	 PATTERNS = {}, -- filled in below
};
COMM.PATTERNS = {
	 SEARCH = {
			BASIC_HEADER     = "^"..COMM.HEADER.PREFIX..COMM.SEP.ENTRY.."(%d+)".."([^"..COMM.SEP.HEADER.."]*)"..COMM.SEP.HEADER,
			HEADERv2         = "^"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)",
			DATA_HEADER      = "^".."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY,
			HASH_PROP_OLD    = "^".."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)$",
			HASH_PROPAGATION = "^".."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(%d+)$",
			DATA_REQUEST     = "^".."(%d+)"..COMM.SEP.ENTRY.."(%d+)"..COMM.SEP.ENTRY.."(.+)$", -- last field: utf8 char names etc
	 },
	 BUILD = {
			FRAME_HEADER = COMM.HEADER.PREFIX..COMM.SEP.ENTRY.."%s"..COMM.SEP.ENTRY.."%d"..COMM.SEP.ENTRY.."%d"..COMM.SEP.HEADER,
			DATA_SECTION = "%d"..COMM.SEP.ENTRY.."%d"..COMM.SEP.ENTRY.."%s",
			ENTRY_SET    = "%s"..COMM.SEP.FIELD.."%s"..COMM.SEP.FIELD.."%s",
	 },
};

local PATTERN = {
	 PLAYER = {
			-- Plain player link without chat id or channel.
			-- May include color definitions etc in second part, so anti-greedy matching
			-- %1 = first part of player link
			-- %2 = sender
			-- %3 = name displayed for link, possibly (but not necessarily) including colors, brackets, etc.
			-- %4 = remaining link. Should be "|h", but who cares?
			PLAIN_LINK    = "(|Hplayer:([^|:]+)|h)(.-)(|h)",

			-- Player link including chat id, channel, etc
			-- May include color definitions etc in second part, so anti-greedy matching
			-- %1 = first part of player link
			-- %2 = sender
			-- %3 = channel, e.g. "WHISPER", "GUILD", etc.
			-- %4 = name displayed for link, possibly (but not necessarily) including colors, brackets, etc.
			-- %5 = remaining link. Should be "|h", but who cares?
			CHANNEL_LINK  = "(|Hplayer:([^|:]+):%d+:(%u+)[^|]*|h)(.-)(|h)",

			-- Console output when player goes offline
			-- %1 = char name
			GOING_OFFLINE = L["ANNOUNCE_OFFLINE"], -- see in locales for pattern, has to be localized
	 },
};

--[[ Local constants ]]
local CONST = {
	 DB_VERSION        =  3,  -- Database format
	 MAXLENGTH_CHAR    = 12,  -- Maximum char name length, defined by Blizzard
	 MINLENGTH_CHAR    =  2,  -- Minimum char name length, defined by Blizzard
	 MAXLENGTH_BUDDY   = 16,  -- Maximum buddy name length, defined by me
	 MAXLENGTH_PATTERN = 25,

	 ERROR = {  -- bias numbers for error messages
			ARG  = 1000, -- next free error number: 47
			VAL  = 3000, -- next free error number: 13
			MISC = 5000, -- next free error number: 13
	 },
}
--[[ Local configuration table for use with AceConfig-3.0 ]]
BuddyName.optionsTable = {
	 type = "group",
	 name = "BuddyName",
	 handler = BuddyName,
	 args = {
			active = {
				 order = 1,
				 type = "toggle",
				 name = L["CFG_ACTIVATE_BN"],
				 width = "full",
				 get = "CfgGetActive",
				 set = "CfgSetActive",
			},
			guildsync = {
				 order = 2,
				 type = "toggle",
				 name = L["CFG_ACTIVATE_GUILDSYNC"],
				 width = "full",
				 disabled = function() return (GetGuildInfo("player")==nil) end,
				 get = "CfgGetSync",
				 set = "CfgSetSync",
			},
			guildsyncmsg = {
				 order = 3,
				 type = "toggle",
				 name = L["CFG_GUILDSYNC_INFO"],
				 width = "full",
				 disabled = function() return (not BuddyName:CfgGetSync()) end,
				 get = "CfgGetSyncMsg",
				 set = "CfgSetSyncMsg",
			},
			guildsyncoverwrite = {
				 order = 4,
				 type = "toggle",
				 name = L["CFG_GUILDSYNC_OVERWRITE"],
				 width = "full",
				 disabled = function() return (not BuddyName:CfgGetSync()) end,
				 get = "CfgGetSyncOverwrite",
				 set = "CfgSetSyncOverwrite",
			},
			displayownname = {
				 order = 6,
				 type = "toggle",
				 name = L["CFG_DISPLAY_OWN_NAME"],
				 width = "full",
				 get = "CfgGetDisplayOwnName",
				 set = "CfgSetDisplayOwnName",
			},
			userPattern = {
				 order = 8,
				 type = "input",
				 name = L["CFG_PATTERN_TITLE"],
				 desc = L["CFG_PATTERN_DESC"],
				 width = "normal",
				 get = "CfgGetUserPattern",
				 set = "CfgSetUserPattern",
				 validate = "CfgValidateUserPattern",
				 multiline = false,
			},
			channels = {
				 order = 10,
				 type = "multiselect",
				 name = L["CFG_CHAT_CHANNELS"],
				 desc = L["CFG_CHAT_CHANNELS_DESC"],
				 disabled = function () return (not BuddyName:CfgGetActive()) end,
				 values = {
						["BATTLEGROUND"] = L["CFG_BATTLEGROUND"],
						["CHANNEL"]      = L["CFG_CHANNEL"],
						["GUILD"]        = L["CFG_GUILD"],
						["PARTY"]        = L["CFG_PARTY"],
						["RAID"]         = L["CFG_RAID"],
						["WHISPER"]      = L["CFG_WHISPER"],
						["YELL"]         = L["CFG_YELL"],
						["OTHER"]        = L["CFG_OTHER"],
				 },
				 set = "CfgSetChannel",
				 get = "CfgGetChannel",
			},
	 },
};

local cfgDefaults = {
	 char = {
			cfg = {
				 active  = true, -- whether BN is active
				 channel = {     -- in which channels to perfom name substitution
						['*'] = true,
				 },
				 displayOwnName = true,
			},
	 },
	 realm = {
			cfg = {
				 guild = {
						['*'] = {
							 sync          = false, -- whether sync is activated by default
							 syncMsg       = false, -- whether to display additional info in chat
							 syncOverwrite = false, -- whether existing entries may be overwritten
						},
				 },
			},
	 },
	 global = {
			cfg = {
				 userPattern = "%char:%buddy",
			},
	 },
};

--[[ Local variables ]]
local activeFlag = false; -- Is this addon active? Modified by OnEnable and OnDisable
local BN_EventFrame = nil; -- Used to register events and process them. Created in OnInitialize
local currentRealm = nil;
local storedTimers = {}; -- table for storing active timers
local warnedAboutUpdate = false; -- whether the user has been warned about an existing update this session
local debugMode = false; -- whether to output verbose debugging messages
local guildHash = nil;

local comm = {
	 hash = {};
	 diff = {};
};

--[[ -- -------------------------------- -- ]]
--[[ -- Basic functions, no dependenices -- ]]
--[[ -- -------------------------------- -- ]]

--[[ Outputs a string to the default chat frame with BuddyName(debug)
--   prefix if debug flag is set.
--   Returns: nil
--]]
local function debug(...)
	 if (debugMode==true) then
			local msg="|cFF33FF99BuddyName (|cFFFF0000Debug|r):|r "..strjoin(":", tostringall(...));
			print(msg);
	 end
end

--[[ Outputs a string to the default chat frame with comm colors
--   Returns: nil
--]]
local function commPrint(...)
	 if (BuddyName:CfgGetSyncMsg()) then
			local msg="|cFFFF8844BuddyName: "..strjoin(" ", tostringall(...)).."|r";
			print(msg);
	 end
end

--[[ Outputs a status string to the default chat frame with status colors
--   Returns: nil
--]]
local function statPrint(...)
	 local msg="|cFF33FF99BuddyName:|r "..strjoin(" ", tostringall(...));
	 print(msg);
end

--[[ Used to check correct arguments for functions.
--   If the first argument does not equal true, then
--   an error is raised. Also, the real value of the
--   argument, as provided by the second function
--   argument, is printed in the error message.
--   Returns: nil
--]]
local function argAssert(errNum, argNum, boolValue, realValue)
	 if (boolValue~=true) then
			errNum = errNum + CONST.ERROR.ARG;
			local errPrefix = string.format("Error %d:%d:%d", BuddyNameBuild, errNum, argNum);
			statPrint("|cFFFF0000"..errPrefix.."|r");
			error(string.format(errPrefix..": Invalid function argument #%d: %s", argNum, tostring(realValue)), 3);
	 end
end

--[[ Same as argAssert, but for any variable values instead of
--   just function arguments.
--   Returns: nil
--]]
local function valAssert(errNum, boolValue, realValue)
	 if (boolValue~=true) then
			errNum = errNum + CONST.ERROR.VAL;
			local errPrefix = string.format("Error %d:%d", BuddyNameBuild, errNum);
			statPrint("|cFFFF0000"..errPrefix.."|r");
			error(string.format(errPrefix..": Invalid variable value: %s", tostring(realValue)), 2);
	 end
end

--[[ Internal error call that uses the official error function, but also adds
--   a given error number as prefix.
--   Returns: nil
--]]
local function Error(errNum, errMsg, errLevel)
	 errNum = errNum + CONST.ERROR.MISC;
	 local errPrefix = string.format("Error %d:%d", BuddyNameBuild, errNum);
	 statPrint("|cFFFF0000"..errPrefix.."|r");
	 error(errPrefix..": "..errMsg, errLevel+1);
end

local function IsBool(arg)
	 return (type(arg)=="boolean");
end

local function IsString(arg)
	 return (type(arg)=="string");
end

local function IsNonEmptyString(arg)
	 return (type(arg)=="string" and arg~="");
end

local function IsNumber(arg)
	 return (type(arg)=="number");
end

local function IsTable(arg)
	 return (type(arg)=="table");
end

local function IsFunc(arg)
	 return (type(arg)=="function");
end

--[[ Creates a new timestamp for the current time and date
--   Returns:
--     1. String: timestamp in format YYYYMMDDHHmm
--]]
function BuddyName:GetTimestamp()
	 local _,month,day,year = CalendarGetDate();
	 local hour,minute = GetGameTime();

	 local timestamp = string.format("%04d%02d%02d%02d%02d", year, month, day, hour, minute);

	 return timestamp;
end

--[[ Concatenates an arbitrary number of arguments to one string.
--   Only used for debugging purposes.
--   Returns:
--     1. String: Concatenated version of all supplied arguments
--]]
local function AllToString(...)
	 return strjoin(" ", tostringall(...));
end

--[[ Checks if currently active character is in a guild, and if so,
--   returns the name of the guild.
--   Returns:
--     1. String: Name of players guild, if player is in a guild, nil else
--]]
local function GetPlayerGuild()
	 local guild = GetGuildInfo("player");
	 return guild;
end

--[[ This function takes a char name, reformat and then returns it.
--   At the moment only the first character of the name is checked
--   and, if necessary, converted to uppercase, while all other
--   characters will be converted to lowercase. Will most likely
--   not work with names starting with a UTF-8 character; on the
--   other side, it should not do any harm there as well, so thats
--   ok for the moment.
--   Returns:
--     1. String: Reformated character name
--]]
function BuddyName:FormatCharName(name)
	 argAssert(46, 1, IsNonEmptyString(name), name);

	 local firstChar = string.sub(name, 1, 1);
	 local remainder = string.sub(name, 2);

	 firstChar = string.upper(firstChar);
	 remainder = string.lower(remainder);

	 return (firstChar..remainder);
end

--[[ Checks a given character name for validity. Checks are
--   performed regarding name length and whether the name
--   contains illegal characters. If the name is not valid,
--   an error message is written to the console, if silent flag
--   is not set.
--   However, this function does NOT check whether a player with
--   this name is online or even exists.
--   Returns:
--     1. Boolean: true if character name is valid; false else
--]]
function BuddyName:IsValidCharName(name, silent)
	 argAssert(1, 1, name~=nil, name);

	 -- check name length
	 if (strlenutf8(name)>CONST.MAXLENGTH_CHAR) then
			if (not silent) then
				 statPrint(string.format(L["CHARNAME_TOO_LONG"], strlenutf8(name), CONST.MAXLENGTH_CHAR));
			end
			return false;
	 end

	 if (strlenutf8(name)<CONST.MINLENGTH_CHAR) then
			if (not silent) then
				 statPrint(string.format(L["CHARNAME_TOO_SHORT"], strlenutf8(name), CONST.MINLENGTH_CHAR));
			end
			return false;
	 end

	 -- search for illegal characters
	 local pstart,pend,pchar = string.find(name,"([|;:_ \t\n\r%z%d%.%(%)%%%+%-%*%?%[%]%^%$])");
	 if (pstart) then
			if (not silent) then
				 statPrint(string.format(L["CHARNAME_ILLEGAL_CHAR"], pchar));
			end
			return false;
	 end

	 return true;
end

--[[ Checks a given buddy name for validity. Checks are
--   performed regarding name length and whether the name
--   contains illegal characters. If the name is not valid,
--   an error message is written to the console, if silent flag
--   is not set. Names with length 0 are treated as valid.
--   Returns:
--     1. Boolean: true if buddy name is valid; false else
--]]
function BuddyName:IsValidBuddyName(name, silent)
	 argAssert(2, 1, name~=nil, name);

	 -- search for illegal characters
	 local pstart,pend,pchar = string.find(name,"([\t\n\r%z%^%$])");
	 if (pstart) then
			if (not silent) then
				 statPrint(string.format(L["BUDDYNAME_ILLEGAL_CHAR"], pchar));
			end
			return false;
	 end

	 -- TODO: When doing name length checking, count without color codes etc.

	 -- check name length
	 if (strlenutf8(name)>CONST.MAXLENGTH_BUDDY) then
			if (not silent) then
				 statPrint(string.format(L["BUDDYNAME_TOO_LONG"], strlenutf8(name), CONST.MAXLENGTH_BUDDY));
			end
			return false;
	 end

	 return true;
end

--[[ Generic function to test if a given value lies within a given
--   range.
--   Returns:
--     1. Boolean: true if value is in range, false else.
--        Also false if the value to check is nil. Error if one
--        of the boundaries is nil.
--]]
local function IsInDataRange(value, lower, upper)
	 argAssert(3, 1, lower~=nil, lower);
	 argAssert(3, 2, upper~=nil, upper);

	 -- range check
	 if (value==nil or value < lower or value > upper) then
			return false;
	 end

	 return true;
end

--[[ Checks if compression type is valid.
--   Returns:
--     1. Boolean: true if valid, false else
--]]
local function IsValidCompression(compression)
	 return IsInDataRange(compression, COMM.COMPRESSION.NONE, COMM.COMPRESSION.LIBCOMPRESS);
end

--[[ Checks if section type is valid.
--   Returns:
--     1. Boolean: true if valid, false else
--]]
local function IsValidSectionType(sectionType)
	 return IsInDataRange(sectionType, COMM.SECTION_TYPE.ENTRY_SET, COMM.SECTION_TYPE.DATA_REQUEST);
end

--[[ Checks if entry set type is valid.
--   Returns:
--     1. Boolean: true if valid, false else
--]]
local function IsValidEntrySetType(entrySetType)
	 return IsInDataRange(entrySetType, COMM.ENTRY_SET_TYPE.FULL_DUMP, COMM.ENTRY_SET_TYPE.SINGLE_ENTRY);
end

--[[ Checks if data domain is valid.
--   Returns:
--     1. Boolean: true if valid, false else
--]]
local function IsValidDataDomain(dataDomain)
	 return IsInDataRange(dataDomain, COMM.DATA_DOMAIN.REALM, COMM.DATA_DOMAIN.GUILD_DIFF);
end

--[[ Checks if response method is valid.
--   Returns:
--     1. Boolean: true if valid, false else
--]]
local function IsValidResponseMethod(responseMethod)
	 return IsInDataRange(responseMethod, COMM.RESPONSE_METHOD.GUILD, COMM.RESPONSE_METHOD.WHISPER);
end

--[[ Checks if response flag is valid.
--   Returns:
--     1. Boolean: true if valid, false else
--]]
local function IsValidResponseFlag(responseFlag)
	 return IsInDataRange(responseFlag, COMM.IS_NO_RESPONSE, COMM.IS_RESPONSE);
end

--[[ Compresses a given string with the given compression method.
--   Returns:
--     1. String: String compressed with given method; if no valid
--        compression type was provided, nil is returned.
--]]
function BuddyName:Compress(myString, compression)
	 argAssert(4, 1, IsNonEmptyString(myString), myString);
	 argAssert(4, 2, IsValidCompression(compression), compression);

	 if (compression==COMM.COMPRESSION.NONE) then
			return myString;
	 elseif (compression==COMM.COMPRESSION.LIBCOMPRESS) then
			myString = libComp:Compress(myString);
			myString = libCompEnc:Encode(myString);
			return myString;
	 end

	 return nil;
end

--[[ Decompresses a given string that is compressed with the given
--   compression method.
--   Returns:
--     1. String: Decompressed string; nil if no valid compression typ
--        is provided.
--]]
function BuddyName:Decompress(myString, compression)
	 argAssert(5, 1, IsNonEmptyString(myString), myString);
	 argAssert(5, 2, IsValidCompression(compression), compression);

	 if (compression==COMM.COMPRESSION.NONE) then
			return myString;
	 elseif (compression==COMM.COMPRESSION.LIBCOMPRESS) then
			myString = libCompEnc:Decode(myString);
			myString = libComp:Decompress(myString);
			return myString;
	 end

	 return nil;
end

--[[ Sets and stores a timer that will call the given function when it
--   expires. When a second timer is set on the same callback function,
--   it will overwrite the first one. Delay is given as seconds.
--   Returns: nil
--]]
function BuddyName:SetTimer(callback, delay)
	 argAssert(6, 1, IsNonEmptyString(callback), callback);
	 argAssert(6, 2, IsNumber(delay) and delay>=0, delay);

	 self:ResetTimer(callback); -- cancel timer if already exists
	 storedTimers[callback] = self:ScheduleTimer(callback, delay);
end

--[[ Resets the timer for the given callback function, if such
--   a timer exists. Else, do nothing.
--   Returns: nil
--]]
function BuddyName:ResetTimer(callback)
	 argAssert(7, 1, IsNonEmptyString(callback), callback);

	 if (storedTimers[callback]) then
			-- no error message if timer does not exist
			self:CancelTimer(storedTimers[callback], true);
	 end
end

--[[ -- ------------------------ -- ]]
--[[ -- Config-related functions -- ]]
--[[ -- ------------------------ -- ]]

--[[ Configures in which chat channels buddy names will be added.
--   The key refers to the source as used in the prefix of
--   AddMessage(). One other key, "OTHER", is used for things
--   like going on-/offline, or achievements, etc.
--   Returns: nil
--]]
function BuddyName:CfgSetChannel(info,key,value)
	 argAssert(8, 2, IsNonEmptyString(key), key);
	 argAssert(8, 3, IsBool(value), value);

	 self.db.char.cfg.channel[key] = value;
end

--[[ Retrieves the current configuration status for a chat channel,
--   i.e. whether to add buddy names in this channel (true) or not
--   (false).
--   Returns:
--     1. Boolean: true if buddy names should be displayed, false else
--]]
function BuddyName:CfgGetChannel(info,key)
	 argAssert(9, 2, IsNonEmptyString(key), key);

	 return self.db.char.cfg.channel[key];
end

--[[ Sets a new configuration status for guild synchronization.
--   The setting is stored on a per-guild basis in the local DB.
--   If the player is not in a guild, do nothing.
--   Returns: nil
--]]
function BuddyName:CfgSetSync(info,value)
	 argAssert(10, 2, IsBool(value), value);

	 local guildName, _, _ = GetGuildInfo("player");

	 if (guildName==nil) then
			return;
	 end

	 self.db.realm.cfg.guild[guildName].sync = value;

	 -- if sync is activated, initiate guild sync process
	 if (value==true) then
			self:SetTimer("InitiateGuildSync", 10);
	 end
end

--[[ Retreives the current configuration status for guild synchronization
--   from the DB.
--   Returns:
--     1. Boolean: true if sync is activated and player is member of
--        a guild; false else.
--]]
function BuddyName:CfgGetSync(info)
	 local guildName, _, _ = GetGuildInfo("player");

	 if (guildName==nil) then
			return false;
	 end

	 return self.db.realm.cfg.guild[guildName].sync;
end

--[[ Sets the flag whether new entries received via
--   synchronization may overwrite existing local entries.
--   The setting is stored on a per-guild basis in the local DB.
--   Returns: nil
--]]
function BuddyName:CfgSetSyncOverwrite(info,value)
	 argAssert(11, 2, IsBool(value), value);

	 local guildName, _, _ = GetGuildInfo("player");

	 if (guildName==nil) then
			return;
	 end

	 self.db.realm.cfg.guild[guildName].syncOverwrite = value;
end

--[[ Returns whether new entries received via synchronization
--   may overwrite existing local entries.
--   Returns:
--     1. Boolean: true if existing entries may be
--        overwritten by guild synchronization.
--]]
function BuddyName:CfgGetSyncOverwrite(info)
	 local guildName, _, _ = GetGuildInfo("player");

	 if (guildName==nil) then
			return false;
	 end

	 return self.db.realm.cfg.guild[guildName].syncOverwrite;
end

--[[ Sets a new configuration status for displaying information
--   about guild synchronization in chat.
--   The setting is stored on a per-guild basis in the local DB.
--   Returns: nil
--]]
function BuddyName:CfgSetSyncMsg(info,value)
	 argAssert(12, 2, IsBool(value), value);

	 local guildName, _, _ = GetGuildInfo("player");

	 if (guildName==nil) then
			return;
	 end

	 self.db.realm.cfg.guild[guildName].syncMsg = value;
end

--[[ Retreives the current configuration status for displaying
--   information about guild synchronization in chat.
--   Returns:
--     1. Boolean: true if display of sync messages is activated
--        and player is member of a guild; false else.
--]]
function BuddyName:CfgGetSyncMsg(info)
	 local guildName, _, _ = GetGuildInfo("player");

	 if (guildName==nil) then
			return false;
	 end

	 return self.db.realm.cfg.guild[guildName].syncMsg;
end

--[[ Sets a new configuration status for the "display own name"
--   flag.
--   Returns: nil
--]]
function BuddyName:CfgSetDisplayOwnName(info,value)
	 argAssert(13, 2, IsBool(value), value);

	 self.db.char.cfg.displayOwnName = value;
end

--[[ Retreives the current configuration status for the "display your
--   own name" flag from the DB.
--   Returns:
--     1. Boolean: true if own name should be displayed; false else
--]]
function BuddyName:CfgGetDisplayOwnName(info)
	 return self.db.char.cfg.displayOwnName;
end

--[[ Sets a new user-defined pattern to display char names
--   and buddy names in chat.
--   Returns: nil
--]]
function BuddyName:CfgSetUserPattern(info, value)
	 argAssert(14, 2, IsString(value), value);

	 -- Remove leading and trailing whitespace
	 value = string.trim(value);

	 self.db.global.cfg.userPattern = value;
end

--[[ Returns the user-defined pattern to display char names
--   and buddy names in chat.
--   Returns:
--     1. String: User pattern
--]]
function BuddyName:CfgGetUserPattern(info)
	 return self.db.global.cfg.userPattern;
end

--[[ Checks whether a user-defined pattern to display char
--   names and buddy names in chat is valid
--   Returns:
--     1. Boolean: true if pattern is valid; false else
--]]
function BuddyName:CfgValidateUserPattern(info, value)
	 argAssert(15, 2, IsString(value), value);

	 -- Remove leading and trailing whitespace
	 value = string.trim(value);

	 -- TODO: Test for other illegal chars.
	 -- TODO: Test if pattern includes at least %char or %buddy

	 if (strlenutf8(value)>CONST.MAXLENGTH_PATTERN) then
			self:Print(string.format(L["CFG_PATTERN_TOO_LONG"], CONST.MAXLENGTH_PATTERN));
			return false;
	 end

	 if (strlenutf8(value)==0) then
			self:Print(L["CFG_PATTERN_EMPTY"]);
			return false;
	 end

	 return true;
end

--[[ Sets the active status flag in the configuration.
--   This flag is saved on a per-character base.
--   Depending on the value, the Enable() or Disable()
--   functions are called.
--   Returns: nil
--]]
function BuddyName:CfgSetActive(info, value)
	 argAssert(16, 2, IsBool(value), value);

	 self.db.char.cfg.active = value;
	 if (value==true) then
			self:Enable();
	 else
			self:Disable();
	 end
end

--[[ Reads and returns the current active status flag
--   in the configuration for this character.
--   Returns:
--     1. Boolean: true if config says BN is active, false else
--]]
function BuddyName:CfgGetActive(...)
	 return self.db.char.cfg.active;
end

--[[ Sets the current active status for this addon.
--   This status is not saved in the config. This
--   function should only be called by OnEnable()
--   and OnDisable().
--   Returns: nil
--]]
function BuddyName:SetActive(value)
	 argAssert(17, 1, IsBool(value), value);

	 if (value~=nil) then
			activeFlag = value;
	 end
end

--[[ Returns the current active status for this addon.
--   Returns:
--     1. true if BN is enabled at the moment, false else
--]]
function BuddyName:IsActive()
	 return activeFlag;
end

--[[ -- -------------------------- -- ]]
--[[ -- Database-related functions -- ]]
--[[ -- -------------------------- -- ]]

--[[ Checks whether the given charname has a buddy name in our DB. The second flag
--   defaults to false and controls whether a true is returned for chars with
--   buddy name = "".
--   Returns:
--     1. Boolean: true if author was found in DB, false else
--     2. String: Buddy name of charname, or nil if it doesn't exist or equals empty string
--     3. String: Timestamp of buddy entry, or nil if it doesn't exist
--]]
function BuddyName:IsInDB(charname, alsoEmptyBuddys)
	 argAssert(18, 1, IsNonEmptyString(charname), charname);
	 argAssert(18, 2, alsoEmptyBuddys==nil or IsBool(alsoEmptyBuddys), alsoEmptyBuddys);

	 charname=self:FormatCharName(charname);

	 -- check if ANY entry exists for this char
	 if (not self.db.realm.char[charname]) then
			return false,nil,nil;
	 end

	 -- check if entry with empty buddy name exists
	 if (self.db.realm.char[charname].buddy~="" or alsoEmptyBuddys)
	 then
			return true,self.db.realm.char[charname].buddy,self.db.realm.char[charname].ts;
	 end

	 return false,nil,nil;
end

--[[ Adds a new entry to the DB or updates an existing one.
--   If a timestamp is provided, then this will be used.
--   Returns:
--     1. Boolean: true if successul, false else
--]]
function BuddyName:AddToDB(charname, buddyname, timestamp)
	 argAssert(19, 1, charname~=nil, charname);
	 argAssert(19, 2, buddyname~=nil, buddyname);
	 argAssert(19, 3, timestamp==nil or IsNonEmptyString(timestamp), timestamp);

	 if (self:IsValidCharName(charname)==false or self:IsValidBuddyName(buddyname)==false) then
			return false;
	 end

	 charname=self:FormatCharName(charname);

	 if (not timestamp) then
			timestamp = self:GetTimestamp();
	 end

	 self.db.realm.char[charname]= {
			buddy = buddyname,
			guild = "",
			ts    = timestamp,
	 }

	 self:InvalidateGuildHash();

	 return true
end

--[[ Takes a given entry and updates in the DB if necessary. If entry
--   does not yet exist in DB, then add. If entry exists, sync
--   operations may update existing entries, and the timestamp of the
--   entry in the DB is older than the one provided, the entry is
--   updated.
--   Returns:
--     1. Boolean: true if updated, false else
--     2. Boolean: true if new entry, false if existing entry was updated
--]]
-- TODO: Update timestamp to newer TS if chars equal and buddynames equal.
-- In this case, ignore syncOverwrite setting
function BuddyName:UpdateIfNewer(charname, buddyname, timestamp)
	 argAssert(20, 1, IsNonEmptyString(charname), charname);
	 argAssert(20, 2, buddyname~=nil, buddyname);
	 argAssert(20, 3, IsNonEmptyString(timestamp), timestamp);

	 local inDB, buddy, ts = self:IsInDB(charname, true);

	 if (not inDB) then -- entry does not exist at all
			debug(string.format("Added %s => %s", charname, buddyname));
			local result = BuddyName:AddToDB(charname,buddyname,timestamp);
			return result, true;
	 elseif (self:CfgGetSyncOverwrite()==true) then
			if (ts < timestamp or -- received timestamp is newer
	        (ts==timestamp and buddyname~=buddy)) -- solve other conflicts
			then
				 debug(string.format("Updated %s => %s", charname, buddyname));
				 local result = BuddyName:AddToDB(charname,buddyname,timestamp);
				 return result, false;
			end
	 end

	 return false, false;
end

--[[ Integrate given entry set into local DB. Local entries are only
--   overwritten if other timestamp is newer, as checked by the
--   UpdateIfNewer() function
--   Returns:
--     1. Decimal: Number of added entries
--     2. Decimal: Number of updated entries
--]]
function BuddyName:IntegrateEntrySet(entrySet)
	 argAssert(21, 1, IsTable(entrySet), entrySet);

	 local addCounter = 0;
	 local updateCounter = 0;
	 local result, newEntry;

	 for p,v in pairs(entrySet) do
			result, newEntry = BuddyName:UpdateIfNewer(p, v.buddy, v.ts);
			if (result==true) then
				 if (newEntry == true) then
						addCounter = addCounter + 1;
				 else
						updateCounter = updateCounter+1;
				 end
			end
	 end

	 return addCounter, updateCounter;
end

--[[ Deletes an entry from the DB if it exists. Depending on the deleteCompletely flag,
--   an existing entry may either be deleted completely (true) or just set to "" (false).
--   The latter leaves the actual entry intact, resulting in its spreading over the guild
--   synchronization mechanism. Therefore, it is possible to delete this entry from other
--   guild members DBs as well.
--   Returns:
--     1. Boolean: true if deletion was successful, false else
--     2. String: buddy name if deletion was successful, nil else (e.g. char was not in DB)
--]]
function BuddyName:DeleteFromDB(charname, deleteCompletely)
	 argAssert(22, 1, IsNonEmptyString(charname), charname);
	 argAssert(22, 2, deleteCompletely==nil or IsBool(deleteCompletely), deleteCompletely);

	 charname=self:FormatCharName(charname);

	 -- is this char in the DB?
	 if (not self.db.realm.char[charname]) then
			return false,nil
	 end

	 -- perform deletion and set return value
	 local buddyname = self.db.realm.char[charname].buddy
	 if (deleteCompletely==true) then
			self.db.realm.char[charname]=nil
			self:InvalidateGuildHash();
	 else
			self:AddToDB(charname, "", nil);
	 end

	 return true,buddyname
end

--[[ Can be used to fix known errors in the DB.
--   Returns: nil
--]]
function BuddyName:FixDB(patch)
	 argAssert(23, 1, IsNumber(patch), patch);

	 -- TODO: also check DB for entries missing required items. For each DBv

	 -- Fix broken entries from RealID experiment
	 -- should only be necessary for DBv2
	 -- TODO: Don't use IsInDB etc, may break during DB structure update.
	 -- TODO: Instead use raw structures as used in DBv2 and DBv3.
	 if (patch==1) then
			-- remove all char names and buddy names that are not valid. May have happend during the RealID test phase
			for p,_ in pairs(self.db.realm.char) do
				 local inDB, buddy, _ = self:IsInDB(p, true);
				 if (self:IsValidCharName(p)==false or (inDB==true and self:IsValidBuddyName(buddy)==false)) then
						self.db.realm.char[p] = nil;
						statPrint(L["REMOVE_BROKEN_FROM_DB"]);
				 end
			end
			return;
	 end

	 if (patch==2) then
			if (type(self.db.realm.cfg.guild[guildName])=="boolean") then
				 self.db.realm.cfg.guild[guildName] = nil;
			end
			return;
	 end

	 -- convert entries in DB where the char name may not be right due to
	 -- uppercase and lowercase. Convert and, if necessary, remove entries
	 if (patch==3) then
			for p,v in pairs(self.db.realm.char) do
				 local formattedName = self:FormatCharName(p);
				 if (p~=formattedName) then
						if (not self:IsInDB(formattedName, true)) then
							 -- move entry to new name format
							 self.db.realm.char[formattedName] = v;
							 self.db.realm.char[p] = nil;
						else
							 -- keep correctly formatted entry, remote other entry
							 self.db.realm.char[p] = nil;
						end
				 end
			end
			return;
	 end

end

--[[ Transforms old DB entries to newer versions
--   Returns: nil
--]]
function BuddyName:UpdateDB()

	 -- DBv1 => DBv2
	 if (self.db.realm.version==1) then
			statPrint(string.format(L["UPDATING_DB"], 2));

			-- perform updates to v2 here
			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p]) then
						self.db.realm.char[p] = {
							 buddy = v,
							 guild = "",
							 timestamp = BuddyName:GetTimestamp(),
						}
				 end
			end

			self.db.realm.version=2
			statPrint(L["UPDATE_OK"])
	 end

	 -- DBv2 => DBv3
	 if (self.db.realm.version==2) then
			statPrint(string.format(L["UPDATING_DB"], 3));

			-- perform update to v3 here
			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p]) then
						v.ts = tostring(v.timestamp);
						v.timestamp = nil;
				 end
			end

			self.db.realm.version=3
			statPrint(L["UPDATE_OK"])
	 end
end

--[[ Invalidates the guild hash.
--   Returns: nil
--]]
function BuddyName:InvalidateGuildHash()
	 guildHash=nil;
	 --debug("Invalidated guild hash");
end

--[[ Returns the current guild hash. If the hash
--   is invalid, e.g. due to changes to the DB,
--   it is calculated again by calling UpdateGuildHash
--   Returns:
--     1. Decimal: Guild hash over complete data,
--        if player is in a guild, nil else
--     2. Decimal: Guild hash over reduced data,
--        if player is in a guild, nil else
--]]
function BuddyName:GetGuildHash()
	 if (guildHash==nil) then
			self:UpdateGuildHash();
	 end
	 --debug("Returning guild hash "..tostring(guildHash));
	 return guildHash.complete, guildHash.simple;
end

--[[ Calculates a deterministic hash over all buddies in the DB that
--   are also members of the chars current guild. This hash is then
--   stored in an addon-global variable. It has to be recalculated
--   when entries in the DB are changed. If the player is not
--   in a guild then the hash is set to nil.
--   Returns: nil
--]]
function BuddyName:UpdateGuildHash()
	 if (not GetPlayerGuild()) then
			guildHash = nil;
			return;
	 end

	 -- copy all guild entries from local DB
	 local guildDB = {};
	 local entry;
	 for p,v in pairs(self.db.realm.char) do -- check complete realm DB
			if (self.db.realm.char[p] and UnitIsInMyGuild(p)) then -- only guildies
				 entry = {};
				 tinsert(entry, p);
				 tinsert(entry, v.buddy);
				 tinsert(entry, v.ts);
				 tinsert(guildDB,entry);
			end
	 end

	 -- sort table according to char names, string comparison
	 table.sort(guildDB, function(a,b) return(a[1]<b[1]) end);

	 -- calculate simple and complete hash for guildDB
	 local crc32Simple   = libComp:fcs32init(); -- crc32 starts as -1
	 local crc32Complete = libComp:fcs32init();
	 for i=1, #guildDB do
			entry = guildDB[i];

			-- complete entry
			crc32Complete = libComp:fcs32update(crc32Complete, entry[1]);
			crc32Complete = libComp:fcs32update(crc32Complete, entry[2]);
			crc32Complete = libComp:fcs32update(crc32Complete, entry[3]);

			-- simplified entry, only char names, and only if buddy != ""
			if (IsNonEmptyString(entry[2])) then
				 crc32Simple = libComp:fcs32update(crc32Complete, entry[1]);
			end
	 end

	 guildHash = {
			simple   = crc32Simple,
			complete = crc32Complete,
	 }
end

--[[ Creates and returns an entry set that reflects the differences
--   between certain entries of the local database and a remote entry.
--   The first argument is the set of remote entries. The second argument
--   can either be another entry set that is used to compare against the
--   first, or can be a function that takes a charname as input and
--   decides whether the according local DB entry should be included for
--   the comparison.
--   Returns:
--     1. Boolean: Flag that indicates wheather the resulting entry set
--        contains any entries (true) or not (false)
--     2. Entry set: Differences between both sets. Empty set if no difference.
--]]
function CreateDiffSet(remoteEntries, arg2)
	 argAssert(24, 1, IsTable(remoteEntries), remoteEntries);
	 argAssert(24, 2, IsTable(arg2) or IsFunc(arg2), arg2);

	 local nonEmpty = false;

	 -- build local entry list, depending on type of 2nd argument
	 local localEntries;
	 if (IsFunc(arg2)) then
			local checkFunc = arg2;
			localEntries = {}; -- start with empty table
			for p,v in pairs(BuddyName.db.realm.char) do
				 if (checkFunc(p)) then -- Do NOT check against ==true, UnitIsInMyGuild returns 1
						localEntries[p] = v;
				 end
			end
			-- debug code block START
			local debugString = "localEntries:";
			for p,_ in pairs(localEntries) do
				 debugString = debugString.." "..p;
			end
			debug(debugString);
			-- debug code block END
	 else
			localEntries = arg2;
	 end

	 -- TODO: Should we distinguish here between different syncOverwrite
	 -- settings? Does this make sense? After all, we don't know what is
	 -- configured at the remote node. Timestamp comparison should work
	 -- in each case, but may contain unnecessary entries...

	 -- compare both sets according to timestamp
	 local resultSet = {};
	 for p,v in pairs(localEntries) do
			if (not remoteEntries[p] or
					localEntries[p].ts > remoteEntries[p].ts or
					(localEntries[p].ts == remoteEntries[p].ts and
					 localEntries[p].buddy ~= remoteEntries[p].buddy))
			then
				 debug("Adding to diff set: "..p.." is "..(v.buddy));
				 resultSet[p] = v;
				 nonEmpty = true;
			end
	 end

	 return nonEmpty, resultSet;
end

--[[ Takes a filter function as argument and checks each entry in the local DB
--   against this function. The resulting list is build from all entries
--   that return true for the filter function. The filter function is given
--   3 strings as arguments: char name, buddy name, timestamp.
--   The resulting list is sorted by charname.
--   Returns:
--     1. Table: Sorted list of entries where the filter function returned true
--]]
function BuddyName:GetSortedList(filterFunc)
	 argAssert(25, 1, IsFunc(filterFunc), filterFunc);

	 local result = {};
	 local entry;

	 -- copy all entries from local DB for which the filter returns true
	 for p,v in pairs(self.db.realm.char) do -- check complete realm DB
			if (filterFunc(p, v.buddy, v.ts)) then
				 entry = {};
				 tinsert(entry, p);
				 tinsert(entry, v.buddy);
				 tinsert(entry, v.ts);
				 tinsert(result, entry);
			end
	 end

	 -- sort table according to char names, string comparison
	 table.sort(result, function(a,b) return(a[1]<b[1]) end);

	 return result;
end

--[[ Takes a filter function as argument and checks each entry in the local DB
--   against this function. The resulting entry set is build from all entries
--   that return true for the filter function. The filter function is given
--   3 strings as arguments: char name, buddy name, timestamp.
--   Returns:
--     1. Table: Entry set from entries where the filter function returned true
--     2. Decimal: Number of entries in resulting entry set
--]]
function BuddyName:GetEntrySet(filterFunc)
	 argAssert(26, 1, IsFunc(filterFunc), filterFunc);

	 local result = {};
	 local numEntries = 0;

	 for p,v in pairs(self.db.realm.char) do
			if (filterFunc(p, v.buddy, v.ts)==true) then
				 result[p]  = v;
				 numEntries = numEntries+1;
			end
	 end

	 return result, numEntries;
end

--[[ Takes a character name as argument and returns the according
--   entry from the local DB as entry set. If the character does
--   not exist in the local DB, a nil value is returned.
--   Returns:
--     1. Table: Entry set consisting of one char, or nil if
--        char does not exist in local DB.
--]]
function BuddyName:GetEntry(charname)
	 argAssert(27, 1, IsNonEmptyString(charname), charname);

	 if (self.db.realm.char[charname]) then
			return { [charname] = self.db.realm.char[charname] };
	 end

	 return nil;
end

--[[ -- -------------- -- ]]
--[[ -- Message output -- ]]
--[[ -- -------------- -- ]]

--[[ Build the string that will replace the charname in case a
--   match was found.
--   Returns:
--     1. String: Replacement string
--]]
local function BuildReplacementString(char, buddy)
	 argAssert(28, 1, IsNonEmptyString(char), char);
	 argAssert(28, 2, IsNonEmptyString(buddy), buddy);

	 local substPattern = BuddyName:CfgGetUserPattern();

	 -- substitute char name
	 substPattern = string.gsub(substPattern, "%%char", char);

	 -- substitute buddy name
	 substPattern = string.gsub(substPattern, "%%buddy", buddy);

	 return substPattern;
end

--[[ Hooked-in version of the AddMessage function.
--   Will check if message contains matching character link,
--   and insert buddy name if match was found.
--   Returns: nil
--]]
function BuddyName:AddMessage(frame, msg, ...)
	 argAssert(29, 2, IsString(msg), msg);

	 -- variables used in string matching below
	 local pstart,pend,linkPart,sender,channel,displayPart,suffixPart;

	 -- match chat channels, including guild, whisper, etc
	 pstart,pend,linkPart,sender,channel,displayPart,suffixPart =
			string.find(msg, PATTERN.PLAYER.CHANNEL_LINK);

	 if (not pstart and self:CfgGetChannel(nil,"OTHER")==true) then -- nothing was found, try next pattern
			-- match people coming online or earning achievements
			pstart,pend,linkPart,sender,displayPart,suffixPart=string.find(msg,PATTERN.PLAYER.PLAIN_LINK); -- check Prat compatibility here
	 end

	 if (pstart) then -- something was found with one of both patterns
			local replace,buddy,_ = self:IsInDB(sender);

			-- channel==nil if 2nd pattern successful, channel=~nil if 1st pattern successful
			if (replace==true and
			   (channel==nil or self:CfgGetChannel(nil,channel)==true) and
			   (self:CfgGetDisplayOwnName() or sender~=GetUnitName("player")))
			then
				 -- split up original message
				 local msgPrefix = string.sub(msg,1,pstart-1); -- everything befor the matched section
				 local msgSuffix = string.sub(msg,pend+1); -- everything after the matched section

				 -- substitute first occurrence of char name in displayPart
				 local substitution=BuildReplacementString(sender,buddy);
				 displayPart=string.gsub(displayPart, sender, substitution, 1);

				 -- rebuild new link
				 local msgLink = linkPart..displayPart..suffixPart;

				 -- rebuild original message, including buddyname
				 msg=msgPrefix..msgLink..msgSuffix;
			end

	 -- both patterns not successful, test for "going offline" message
	 elseif(self:CfgGetChannel(nil,"OTHER")==true) then
			pstart,pend,sender=string.find(msg, PATTERN.PLAYER.GOING_OFFLINE);
			if (pstart) then
				 local replace,buddy,_ = self:IsInDB(sender); -- char in DB?
				 if (replace) then
						local substitution=BuildReplacementString(sender,buddy);
						msg=string.gsub(msg, sender, substitution, 1);
				 end
			end
	 end

	 -- call original AddMessage() function
	 self.hooks[frame]["AddMessage"](frame, msg, ...);
end

--[[ Event Handler that is currently not in use. Perhaps in future version agains.
--   Returns: nil
--]]
local function eventHandler(self, event, ...)
	 if (BuddyName:IsActive() == false) then
			return;
	 end

	 if (registerEvents[event]==true) then
			--print("Registered event of type "..event); -- debug
	 end
end

--[[ -- ----------------- -- ]]
--[[ -- AddOn interaction -- ]]
--[[ -- ----------------- -- ]]

--[[ Process slash commands (unexpected, eh?)
--   Returns: nil
--]]
function BuddyName:SlashProcessor(input)
	 argAssert(30, 1, IsString(input), input);
	 -- 'input' contains whatever follows the slash command

	 -- TODO: respect text contained in "..." or '...', split correctly

	 -- split input string, ignore additional whitespace
	 local tbl = {}
	 for v in string.gmatch(input, "[^ ]+") do
			tinsert(tbl, v)
	 end

	 -- if no additional parameters => print out help text
	 if (#tbl == 0 or (tbl[1]~=nil and tbl[1]=="help")) then
			statPrint(L["HELP"])
			return;
	 end

	 local command = string.lower(tremove(tbl, 1));

	 -- toggle the output of debug statements
	 if (command=="debug") then
			if (debugMode==true) then
				 debugMode=false;
				 statPrint(L["DEBUG_OFF"]);
			else
				 debugMode=true;
				 statPrint(L["DEBUG_ON"]);
			end
			return;
	 end

	 -- Activate additional functionality
	 if (command=="on") then
			if (not self:IsActive()) then
				 self:Enable();
				 statPrint(L["ACTIVATED"]);
			else
				 statPrint(L["ALREADY_ACTIVE"]);
			end
			return
	 end

	 -- Disable additional functionality
	 if (command=="off") then
			if (self:IsActive()) then
				 self:Disable();
				 statPrint(L["DEACTIVATED"]);
			else
				 statPrint(L["NOT_ACTIVE"])
			end
			return
	 end

	 -- stop here if addon is not active
	 if (not self:IsActive()) then
			statPrint(L["NOT_ACTIVE"])
			return;
	 end

	 -- add another person to the list
	 if (command=="add") then
			if (#tbl~=2) then
				 statPrint("Add: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
				 return;
			end
			tbl[1] = self:FormatCharName(tbl[1]);
			local successful = self:AddToDB(tbl[1], tbl[2]);
			if (successful) then
				 statPrint(string.format(L["BUDDY_ADD_OK"].." "..L["CHAR_BUDDY"], tbl[1], tbl[2]));
				 if (self:CfgGetSync()==true and UnitIsInMyGuild(tbl[1])) then

						local entry = self:GetEntry(tbl[1]);

						valAssert(1, IsTable(entry), entry);

						commPrint(L["SYNC_SHARE_SINGLE"]);
						self:BroadcastSingleEntryToGuild(entry);
				 end
			else
				 statPrint(string.format(L["BUDDY_ADD_ERROR"].." "..L["CHAR_BUDDY"], tbl[1], tbl[2]));
			end
			return;
	 end

	 -- delete person from the list
	 if (command=="del" or command=="rm") then
			if (#tbl~=1) then
				 statPrint("Del: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
				 return;
			end
			tbl[1] = self:FormatCharName(tbl[1]);
			local successful,buddyname = self:DeleteFromDB(tbl[1], false)
			if (not successful) then
				 statPrint(string.format(L["BUDDY_DEL_ERROR"], tbl[1]))
				 return;
			end
			statPrint(string.format(L["BUDDY_DEL_OK"].." "..L["CHAR_BUDDY"], tbl[1], buddyname));
			if (self:CfgGetSync()==true and UnitIsInMyGuild(tbl[1])) then

				 local entry = self:GetEntry(tbl[1]); -- see above, buddy name was just set to ""

				 valAssert(2, IsTable(entry), entry);

				 commPrint(L["SYNC_SHARE_SINGLE"]);
				 self:BroadcastSingleEntryToGuild(entry);
			end
			return;
	 end

	 -- list chars from database
	 if (command=="list") then

			-- no further arguments, print complete server list
			local outputList;
			if (#tbl==0) then
				 statPrint(string.format(L["BUDDY_LIST_COMPLETE"],currentRealm));
				 outputList = self:GetSortedList(function(c,b,ts) return true; end);
			elseif (#tbl==1 and tbl[1]=="guild") then
				 local guild = GetPlayerGuild();
				 if (guild==nil) then
						statPrint(L["NOT_IN_GUILD"]);
						return;
				 end
				 statPrint(string.format(L["BUDDY_LIST_GUILD"],guild,currentRealm));
				 outputList = self:GetSortedList(function(c,b,ts) return UnitIsInMyGuild(c) end);
			elseif (#tbl==1) then
				 statPrint(string.format(L["BUDDY_LIST_CONTAINS"],currentRealm, tbl[1]));
				 outputList = self:GetSortedList(function(c,b,ts) return (string.find(c,tbl[1]) or string.find(b,tbl[1])) end);
			else
				 statPrint("List: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
				 return;
			end

			for i=1, #outputList do
				 local entry = outputList[i];
				 if (entry[2] and entry[2]~="") then
						statPrint(string.format(L["CHAR_BUDDY"], entry[1], entry[2]));
				 end
			end

			return;
	 end

	 -- switch on communication properties
	 if (command=="sync") then
			if (self:CfgGetSync()==true) then
				 self:CfgSetSync(_,false);
				 statPrint(L["SYNC_OFF"]);
			else
				 self:CfgSetSync(_,true);
				 statPrint(L["SYNC_ON"]);
			end
			return;
	 end

	 -- exit here if not in debug mode
	 if (debugMode==false) then
			statPrint(string.format(L["UNKNOWN_COMMAND"], command));
			return
	 end

	 -- -------------------------------------- --
	 -- commands only active in debugging mode --
	 -- -------------------------------------- --

	 if (self:CfgGetSync()==false) then
			statPrint(L["SYNC_INACTIVE"]);
			return;
	 end

	 -- to test the communication properties
	 if (command=="bc") then
			statPrint(L["COMM_GUILD_BROADCAST"]);
			self:BroadcastAllToGuild();
			return;
	 end

	 if (command=="hash") then
			statPrint(L["COMM_GUILD_BROADCAST"]);
			local completeHash, simpleHash = self:GetGuildHash();
			self:PropagateHash(COMM.DATA_DOMAIN.GUILD, completeHash, simpleHash, COMM.IS_NO_RESPONSE, "GUILD", nil);
			return;
	 end

end

--[[ -- ------------------------- -- ]]
--[[ -- AddOn chat communication  -- ]]
--[[ -- ------------------------- -- ]]

--[[ Adds a data section header to an existing data section.
--   The section type has to be provided, the section length
--   determined automatically. Returns nil if one of the
--   arguments is nil.
--   Returns:
--     1. Complete data section as string, including section header.
--]]
local function CommEncodeSectionHeader(sectionType, sectionBody)
	 argAssert(31, 1, IsValidSectionType(sectionType), sectionType);
	 argAssert(31, 2, IsString(sectionBody), sectionBody);

	 local length=strlen(sectionBody);

	 local result =
			string.format(COMM.PATTERNS.BUILD.DATA_SECTION,
										sectionType,length,sectionBody);

	 return result;
end

--[[ Takes a string that contains at least one data section.
--   Then automatically determines type and length of section,
--   splits string up accordingly and returns the results. If the
--   data is malformed, e.g. the length indicator is longer than
--   the remaining string, then nil is returned.
--   Returns:
--     1. Data section type as decimal
--     2. Data section body without section header
--     3. String remainder that does not belong to current section
--]]
local function CommDecodeSection(sectionString)
	 argAssert(32, 1, IsNonEmptyString(sectionString), sectionString);

	 -- search for common data section header
	 local pstart,pend,type,length = string.find(sectionString, COMM.PATTERNS.SEARCH.DATA_HEADER);

	 if (pstart) then
			-- cut off data section header
			local remainder=strsub(sectionString,pend+1);

			type=tonumber(type);
			length=tonumber(length);

			local realLength = strlen(remainder);
			if (length==realLength) then
				 return type,remainder,nil;
			elseif (length<realLength) then
				 local sectionBody = strsub(remainder,1,length);
				 remainder = strsub(remainder, length+1);
				 return type,sectionBody,remainder;
			else -- real length is shorter than value from length indicator
				 Error(1, "CommDecodeSection(): Length indicator is wrong: "..length, 1);
			end
	 end
end

--[[ Takes the body of a data section of type ENTRY_SET without the
--   data section header and converts it into a table according to
--   DBv3. Does not insert into DB, just returns the table.
--   Malformed entries are ignored. Nearly no validity checks are
--   done on the data, as long as all fields are present.
--   Returns:
--     1. EntrySetType as number, or nil else
--     2. Table in DBv3 format that includes all entries that
--        were successfully decoded.
--]]
local function CommDecodeEntrySet(esString)
	 argAssert(33, 1, IsNonEmptyString(esString), esString);

	 local entrySet = {};

	 -- split up into single entries according to separator
	 local entries={ strsplit(COMM.SEP.ENTRY, esString) };

	 local esType = tonumber(entries[1]);

	 if (not esType or not IsValidEntrySetType(esType)) then
			Error(2, "CommDecodeEntrySet(): Section with invalid entry set type: "..entries[1], 1);
	 end

	 for i=2, #entries do
			if (IsNonEmptyString(entries[i])) then
				 local entry = { strsplit(COMM.SEP.FIELD, entries[i]); };

				 valAssert(3, #entry==3, entries[i]);

				 local char  = entry[1];
				 local buddy = entry[2];
				 local ts    = entry[3];

				 valAssert(4, IsNonEmptyString(char), char);
				 valAssert(5, tonumber(ts)~=nil, ts);

				 entrySet[char]={
						buddy = buddy,
						guild = "", -- DBv2+3...
						ts    = ts,
				 };
			end
	 end

	 return esType,entrySet;
end

--[[ Decodes a string that represents a frame section of type HASH_PROPAGATION.
--   In case of an invalid section string, an error is raised and nil is returned.
--   Same in case of invalid values in the string.
--   Returns:
--     1. Decimal: Data domain
--     2. Decimal: Complete Hash
--     3. Decimal: Simple hash, if present, else nil
--     4. Decimal: Response flag
--]]
local function CommDecodeHashPropagation(sectionString)
	 argAssert(34, 1, IsNonEmptyString(sectionString), sectionString);

	 -- new frame version
	 local pstart,pend,dataDomain,responseFlag,completeHash,simpleHash = string.find(sectionString, COMM.PATTERNS.SEARCH.HASH_PROPAGATION);

	 if (not pstart) then -- test for old frame format
			pstart,pend,dataDomain,responseFlag,completeHash = string.find(sectionString, COMM.PATTERNS.SEARCH.HASH_PROP_OLD);
	 end

	 -- one of the two formats was found
	 if (pstart) then
			dataDomain   = tonumber(dataDomain);
			responseFlag = tonumber(responseFlag);
			completeHash = tonumber(completeHash);
			if (simpleHash~=nil) then
				 simpleHash = tonumber(simpleHash);
			end

			-- check consistency
			valAssert(6, IsValidDataDomain(dataDomain), dataDomain);
			valAssert(7, IsValidResponseFlag(responseFlag), responseFlag);
			valAssert(8, IsNumber(completeHash), completeHash);
			---valAssert(9, IsNumber(simpleHash), simpleHash); -- include when support for old frame format is removed

			return dataDomain, completeHash, simpleHash, responseFlag;
	 end

	 Error(3, "CommDecodeHashPropagation(): Section string does not match", 1);
end

--[[ Decodes a string that represents a frame section of type DATA_REQUEST.
--   In case of an invalid section string, an error is raised and nil is returned.
--   Same in case of invalid values in the string.
--   Returns:
--     1. Decimal: Data domain
--     2. Decimal: Response method
--     3. String: Receiver
--]]
local function CommDecodeDataRequest(sectionString)
	 argAssert(35, 1, IsNonEmptyString(sectionString), sectionString);

	 local pstart,pend,dataDomain,responseMethod,receiver = string.find(sectionString, COMM.PATTERNS.SEARCH.DATA_REQUEST);

	 if (pstart) then
			dataDomain     = tonumber(dataDomain);
			responseMethod = tonumber(responseMethod);

			-- check consistency
			valAssert(10, IsValidDataDomain(dataDomain), dataDomain);
			valAssert(11, IsValidResponseMethod(responseMethod), responseMethod);
			valAssert(12, BuddyName:IsValidCharName(receiver), receiver);

			return dataDomain, responseMethod, receiver;
	 end

	 Error(4, "CommDecodeDataRequest(): Section string does not match", 1);
end

--[[ Takes a set of DB entries as table formatted according to DBv2
--   and converts it into a section of type ENTRY_SET. The data
--   section header is also added. Nil is returned if table has
--   wrong format.
--   Returns:
--     1. Complete section of type ENTRY_SET, including data
--        section header.
--]]
local function CommEncodeEntrySet(entrySet, esType)
	 argAssert(36, 1, IsTable(entrySet), entrySet);
	 argAssert(36, 2, IsValidEntrySetType(esType), esType);

	 local sectionString = nil;

	 for p,v in pairs(entrySet) do
			if (v.buddy==nil or v.ts==nil) then
				 return nil;
			end

			local entry =
			string.format(COMM.PATTERNS.BUILD.ENTRY_SET, p, v.buddy, v.ts);

			-- Always add separator, except first time
			if (sectionString~=nil) then
				 sectionString=sectionString..COMM.SEP.ENTRY..entry;
			else
				 sectionString=entry;
			end
	 end

	 if (sectionString==nil) then
			return nil;
	 end

	 -- add entry set type to sectionBody
	 sectionString = esType..COMM.SEP.ENTRY..sectionString;

	 -- add header to body
	 sectionString = CommEncodeSectionHeader(COMM.SECTION_TYPE.ENTRY_SET, sectionString);

	 return sectionString;
end

--[[ Takes an entry set as well as some other arguments and then processes
--   the entry set accordingly, i.e. integrates new entries into own DB etc.
--   At the moment only entries from people in the current guild are processed
--   Returns: nil
--]]
local function CommProcessEntrySet(esType, entrySet, channel, sender)
	 argAssert(37, 1, IsValidEntrySetType(esType), esType);
	 argAssert(37, 2, IsTable(entrySet), entrySet);
	 argAssert(37, 3, IsNonEmptyString(channel), channel);
	 argAssert(37, 4, IsNonEmptyString(sender), sender);

	 debug(string.format("CommProcessEntrySet(%s:%s:%s:%s)", tostring(esType), "table", tostring(channel), tostring(sender)));

	 -- only process entry sets from other guild members at the moment
	 if (not UnitIsInMyGuild(sender)) then
			debug("Received entry set from unit that is not in my guild: "..sender);
			return;
	 end

	 -- upon reception of full dump, integrate and build diff set for sender
	 if (esType==COMM.ENTRY_SET_TYPE.FULL_DUMP) then
			local addCounter, updateCounter = BuddyName:IntegrateEntrySet(entrySet);

			if ((addCounter+updateCounter)>0) then
				 print(string.format(L["SYNC_ADD_MULTI"], addCounter, updateCounter));
			end

			-- store diffset for a certain amount of time
			local nonEmpty, diffSet = CreateDiffSet(entrySet, UnitIsInMyGuild);
			if (nonEmpty) then
				 debug("Storing diff set for "..sender);
				 comm.diff[sender] = diffSet;
				 BuddyName:SetTimer("ClearDiffSets", 60);
			else
				 debug("Empty diff set for "..sender);
			end
			return;
	 end

	 -- upon reception of differential dump, integrate and call ProcessDiffHashes
	 if (esType==COMM.ENTRY_SET_TYPE.DIFF_DUMP) then
			local addCounter, updateCounter = BuddyName:IntegrateEntrySet(entrySet);

			if ((addCounter+updateCounter)>0) then
				 commPrint(string.format(L["SYNC_ADD_MULTI"], addCounter, updateCounter));
			end

			return;
	 end

	 -- upon reception of single entry, integrate and do nothing
	 if (esType==COMM.ENTRY_SET_TYPE.SINGLE_ENTRY) then

			if (UnitIsInMyGuild(sender)) then -- TODO: Only sync from guild members at the moment
				 for p,v in pairs(entrySet) do
						if (UnitIsInMyGuild(p)) then -- TODO: Only sync guild members at the moment
							 local result,newEntry = BuddyName:UpdateIfNewer(p, v.buddy, v.ts);
							 if (result==true) then
									if (newEntry) then
										 commPrint(string.format(L["SYNC_ADD_BUDDY"].." "..L["CHAR_BUDDY"], sender, p, v.buddy));
									else
										 commPrint(string.format(L["SYNC_UPD_BUDDY"].." "..L["CHAR_BUDDY"], sender, p, v.buddy));
									end
							 end
						end
				 end
			end

			return;
	 end

end

--[[ Takes the fields contained in a hash propagation section and acts
--   according to the communication protocol.
--   Returns: nil
--]]
local function CommProcessHashPropagation(dataDomain, remoteCompleteHash, remoteSimpleHash, responseFlag, channel, sender)
	 argAssert(38, 1, IsValidDataDomain(dataDomain), dataDomain);
	 argAssert(38, 2, IsNumber(remoteCompleteHash), remoteCompleteHash);
	 argAssert(38, 3, remoteSimpleHash==nil or IsNumber(remoteSimpleHash), remoteSimpleHash);
	 argAssert(38, 4, IsValidResponseFlag(responseFlag), responseFlag);
	 argAssert(38, 5, IsNonEmptyString(channel), channel);
	 argAssert(38, 6, BuddyName:IsValidCharName(sender), sender);

	 debug(string.format("CommProcessHashPropagation(%s:%s:%s:%s:%s:%s)", tostring(dataDomain), tostring(remoteCompleteHash), tostring(remoteSimpleHash), tostring(responseFlag), channel, sender));

	 -- only process guild hashes at the moment
	 if (dataDomain~=COMM.DATA_DOMAIN.GUILD) then
			-- do nothing at the moment
			return;
	 end

	 -- cope with clients that send no simpleHash
	 if (remoteSimpleHash==nil) then
			return;
	 end

	 local completeHash,simpleHash = BuddyName:GetGuildHash();

	 -- only react on differing hashes
	 if ((BuddyName:CfgGetSyncOverwrite()==true and remoteCompleteHash==completeHash) or
			 (BuddyName:CfgGetSyncOverwrite()==false and remoteSimpleHash==simpleHash))
	 then
			return;
	 end

	 if (responseFlag==COMM.IS_NO_RESPONSE) then -- react with response via whisper
			BuddyName:PropagateHash(COMM.DATA_DOMAIN.GUILD, completeHash, simpleHash, COMM.IS_RESPONSE, "WHISPER", sender);
			return;
	 elseif (responseFlag==COMM.IS_RESPONSE) then

			-- choose the correct hash for below
			local remoteHash;
			if (BuddyName:CfgGetSyncOverwrite()==true) then
				 remoteHash = remoteCompleteHash;
			else
				 remoteHash = remoteSimpleHash;
			end

			-- store in table, only one sender per hash, discard others
			if (comm.hash[remoteHash]) then
				 return;
			else
				 comm.hash[remoteHash] = sender;
			end

			-- set timer to 2 secs to broadcast own full dump to guild
			-- TODO: change seconds value to some constant from above
			BuddyName:SetTimer("BroadcastAllToGuild", 2);

			-- set timer to 3 secs to wait for other whispers
			-- TODO: change seconds value to some constant from above
			BuddyName:SetTimer("ProcessDiffHashes", 3);
			return;
	 else -- no valid response flag
			Error(5, "CommProcessHashPropagation(): Found no valid response flag", 1);
			return;
	 end
end

--[[ Clears the list of stored diff sets.
--   Returns: nil
--]]
function BuddyName:ClearDiffSets()
	 debug("ClearDiffSets()");
	 comm.diff = {};
end

--[[ Takes the fields contained in a data request section and acts
--   according to the communication protocol.
--   Returns: nil
--]]
local function CommProcessDataRequest(dataDomain, responseMethod, receiver, sender)
	 argAssert(39, 1, IsValidDataDomain(dataDomain), dataDomain);
	 argAssert(39, 2, IsValidResponseMethod(responseMethod), responseMethod);
	 -- check receiver?
	 argAssert(39, 4, IsNonEmptyString(sender), sender);

	 debug(string.format("CommProcessDataRequest(%s:%s:%s:%s)", tostring(dataDomain), tostring(responseMethod), tostring(receiver), tostring(sender)));

	 -- TODO: check somewhere else that we are indeed identical to the receiver here

	 if (dataDomain==COMM.DATA_DOMAIN.GUILD_DIFF) then
			local responseChannel;
			if (responseMethod==COMM.RESPONSE_METHOD.GUILD) then
				 responseChannel = "GUILD";
			elseif (responseMethod==COMM.RESPONSE_METHOD.WHISPER) then
				 responseChannel = "WHISPER";
			else
				 Error(6, "CommProcessDataRequest(): No valid response channel found", 1);
				 return;
			end

			-- TODO:check if receiver is in current guild
			-- skip because only diff sets of guild members are stored ATM
			-- and also we should be the receiver, therefore we're in the guild

			if (comm.diff[sender]~=nil) then
				 local diffSet = comm.diff[sender];
				 comm.diff[sender] = nil;

				 local sectionString = CommEncodeEntrySet(diffSet, COMM.ENTRY_SET_TYPE.DIFF_DUMP);

				 debug(string.format("Sending DIFF_ES to %s via %s", tostring(sender), tostring(responseChannel)));
				 BuddyName:SendSections(responseChannel, sender, sectionString);
			else
				 debug("CommProcessDataRequest(): Empty diff set for sender "..sender);
			end

			return;
	 end

	 if (dataDomain==COMM.DATA_DOMAIN.GUILD) then
			-- TODO: Send out full guild dump via responseMethod, but only if receiver
			-- is in guild
			debug("CommProcessDataRequest(): Returned because dataDomain==COMM.DATA_DOMAIN.GUILD");
			return;
	 end;

	 -- TODO: domain "realm" still not processed
end

--[[ Takes a complete frame as string, checks the header, and decodes the frames'
--   other data sections according to their format. The result is returned as a
--   table consisting of the header and an ordered list of data sections. In
--   case of a malformed frame, be it in the header or in one of the sections,
--   the protocol field in the header will have the value COMM.UNKNOWN_PROTOCOL.
--   Returns:
--     1. Table = {
--        	 header = table according to CommDecodeFrameHeader()
--        	 sections = {}: indexed table of data sections in order;
--                          each section will itself be a table with two
--                          fields:
--                          type: data section type
--                          body: entries according to the section type
--        }
--]]
local function CommDecodeFrame(frameString)
	 argAssert(40, 1, IsNonEmptyString(frameString), frameString);

	 local frame = {
			header = nil,
			sections = {},
	 };

	 -- search for basic frame header (prefix + protocol number)
	 local hstart,hend,protocol,remainder =
			string.find(frameString, COMM.PATTERNS.SEARCH.BASIC_HEADER);

	 -- no header was found
	 if (not hstart) then
			debug("CommDecodeFrame(): Frame header does not match: "..string.sub(frameString, 1, 7));
			return nil;
	 end

	 protocol=tonumber(protocol);

	 -- Quit if this is not the current protocol
	 if (protocol~=COMM.PROTOCOL) then
			debug("CommDecodeFrame(): Found wrong protocol version: "..protocol);
			-- TODO If protocol version is newer, give a warning once
			return nil;
	 end

	 local rstart,rend,sections,compression =
			string.find(remainder,COMM.PATTERNS.SEARCH.HEADERv2);

	 -- wrong format for protocol v2 header
	 if (not rstart) then
			Error(7, "CommDecodeFrame(): Header says this frame is from protocol v2, but frame does not match", 1);
			return nil;
	 end

	 -- check for valid compression type
	 if (not IsValidCompression(tonumber(compression))) then
			Error(8, "CommDecodeFrame(): Header contains erronous compression flag: "..compression, 1);
			return frame;
	 end

	 frame.header = {
			protocol    = protocol;
			sections    = tonumber(sections),
			compression = tonumber(compression),
	 };

	 -- Use only remaining string without header now
	 local sectionString = strsub(frameString,hend+1);

	 -- decompress if necessary
	 sectionString = BuddyName:Decompress(sectionString,frame.header.compression);
	 if (not sectionString) then
			Error(9, "CommDecodeFrame(): Could not decompress string", 1);
			return;
	 end

	 -- process each section
	 for i=1, frame.header.sections do
			local type,body,remainder = CommDecodeSection(sectionString);

			if (type==nil or body==nil) then
				 debug("Problem decoding section "..i);
				 frame.header.protocol=COMM.UNKNOWN_PROTOCOL;
				 return frame;
			end

			-- check if anything still left although we're processing the last section
			if (i==frame.header.sections and (remainder~=nil and remainder~="")) then
				 debug("Problem decoding frame, additional data found");
				 frame.header.protocol=COMM.UNKNOWN_PROTOCOL;
				 return frame;
			end

			if (type==COMM.SECTION_TYPE.ENTRY_SET) then
				 --debug("Decoding section of type ENTRY_SET");

				 local esType, bodyTbl = CommDecodeEntrySet(body);

				 local section = {
						type     = type,
						esType   = esType,
						entrySet = bodyTbl,
				 };

				 tinsert(frame.sections, section);
			elseif (type==COMM.SECTION_TYPE.HASH_PROPAGATION) then
				 --debug("Decoding section of type HASH_PROPAGATION");

				 local dataDomain, completeHash, simpleHash, responseFlag = CommDecodeHashPropagation(body);

				 local section = {
						type         = type,
						dataDomain   = dataDomain,
						completeHash = completeHash,
						simpleHash   = simpleHash,
						response     = responseFlag,
				 };

				 tinsert(frame.sections, section);
			elseif (type==COMM.SECTION_TYPE.DATA_REQUEST) then
				 --debug("Decoding section of type DATA_REQUEST");

				 local dataDomain,responseMethod,receiver = CommDecodeDataRequest(body);

				 local section = {
						type           = type,
						dataDomain     = dataDomain,
						responseMethod = responseMethod,
						receiver       = receiver,
				 };

				 tinsert(frame.sections, section);
			else
				 Error(10, "Unknown section type found", 1);
				 frame.header.protocol=COMM.UNKNOWN_PROTOCOL;
				 return frame;
			end

			-- prepare next round in for loop
			sectionString = remainder;
	 end

	 return frame;
end

--[[ Creates and returns a HASH_PROPAGATION section for the given
--   data domain. Returns nil if no valid data domain was given.
--   Returns:
--     1. HASH_PROPAGATION section for the given data domain;
--        nil if data domain is invalid
--]]
local function CommEncodeHashPropagation(dataDomain, completeHash, simpleHash, responseFlag)
	 argAssert(41, 1, IsValidDataDomain(dataDomain), dataDomain);
	 argAssert(41, 2, IsNumber(completeHash), completeHash);
	 argAssert(41, 3, IsNumber(simpleHash), simpleHash);
	 argAssert(41, 4, IsValidResponseFlag(responseFlag), responseFlag);

	 local sectionBody = dataDomain..COMM.SEP.ENTRY..tostring(responseFlag)..COMM.SEP.ENTRY..tostring(completeHash)..COMM.SEP.ENTRY..tostring(simpleHash);

	 local section =
	 CommEncodeSectionHeader(COMM.SECTION_TYPE.HASH_PROPAGATION,sectionBody);

	 return section;
end

-- TODO: document me
-- returns string
local function CommEncodeDataRequest(dataDomain, responseMethod, receiver)
	 argAssert(42, 1, IsValidDataDomain(dataDomain), dataDomain);
	 argAssert(42, 2, IsValidResponseMethod(responseMethod), responseMethod);
	 argAssert(42, 3, BuddyName:IsValidCharName(receiver), receiver);

	 local sectionBody =
			dataDomain..COMM.SEP.ENTRY..responseMethod..COMM.SEP.ENTRY..receiver;

	 local section =
			CommEncodeSectionHeader(COMM.SECTION_TYPE.DATA_REQUEST,sectionBody);

	 return section;
end

-- TODO: document me
function BuddyName:InitiateGuildSync()
	 if (self:CfgGetSync()==true) then
			local completeHash, simpleHash = self:GetGuildHash();
			self:PropagateHash(COMM.DATA_DOMAIN.GUILD, completeHash, simpleHash, COMM.IS_NO_RESPONSE, "GUILD", nil);
			self:SetTimer("InitiateGuildSync", 5*60);
	 end
end

-- TODO: document me
function BuddyName:ProcessDiffHashes()
	 debug("ProcessDiffHash");

	 if (comm.hash==nil) then
			return;
	 end

	 -- take one arbitrary entry from hash set
	 for hash,sender in pairs(comm.hash) do
			debug("ProcessDiffHashes: "..hash..":"..sender);
			-- whisper diff data request to remote node
			local section = CommEncodeDataRequest(COMM.DATA_DOMAIN.GUILD_DIFF, COMM.RESPONSE_METHOD.WHISPER, sender);
			debug(string.format("Sending DATA_REQUEST to %s via %s", tostring(sender), "WHISPER"));
			self:SendSections("WHISPER", sender, section);

			-- remove entry from table
			comm.hash[hash] = nil;

			-- set new timer
			BuddyName:SetTimer("ProcessDiffHashes", 2);

			break; -- only one entry, so stop after first run
	 end
end

--[[ Automatically called by AceComm3 if a message with matching prefix
--   was received.
--   Returns: nil
--]]
function BuddyName:OnCommReceived(prefix, frameString, channel, sender)
	 if (self:CfgGetSync()==false) then
			return;
	 end

	 debug("Received frame via "..channel.." ("..sender..")");

	 -- test if frame is from different addon or yourself
	 if (prefix~=COMM.SYNC_PREFIX or sender==GetUnitName("player",false)) then
			return;
	 end

	 -- decode string into tables
	 local frame = CommDecodeFrame(frameString);

	 if (frame==nil) then
			return;
	 end

	 -- ignore everything that is not the current protocol
	 if (frame.header.protocol~=COMM.PROTOCOL) then
			debug("Received frame that is malformed or uses the wrong protocol");
			return;
	 end

	 --debug("Number of sections in frame: "..tostring(frame.header.sections));
	 -- process each section
	 for i=1, frame.header.sections do
			--debug("Processing section "..i);
			local section = frame.sections[i];

			if (section.type==COMM.SECTION_TYPE.ENTRY_SET) then
				 --debug("Found section of type ENTRY_SET ("..tostring(section.esType)..")");
				 CommProcessEntrySet(section.esType, section.entrySet, channel, sender);
			elseif (section.type==COMM.SECTION_TYPE.HASH_PROPAGATION) then
				 --debug("Found section of type HASH_PROPAGATION");
				 CommProcessHashPropagation(section.dataDomain, section.completeHash, section.simpleHash, section.response, channel, sender);
			elseif (section.type==COMM.SECTION_TYPE.DATA_REQUEST) then
				 --debug("Found section of type DATA_REQUEST");
				 -- TODO: only proceed if we are the receiver encoded in the section
				 CommProcessDataRequest(section.dataDomain, section.responseMethod, section.receiver, sender);
			else
				 Error(11, "Found section of unknown type", 1);
			end
	 end
end

-- TODO: document me
function BuddyName:PropagateHash(dataDomain, completeHash, simpleHash, responseFlag, channel, destination)
	 argAssert(43, 1, IsValidDataDomain(dataDomain), dataDomain);
	 argAssert(43, 2, IsNumber(completeHash), completeHash);
	 argAssert(43, 3, IsNumber(simpleHash), simpleHash);
	 argAssert(43, 4, IsValidResponseFlag(responseFlag), responseFlag);
	 argAssert(43, 5, IsNonEmptyString(channel), channel);
	 argAssert(43, 6, destination==nil or IsNonEmptyString(destination), destination);

	 if (not GetPlayerGuild()) then
			return;
	 end

	 local section = CommEncodeHashPropagation(dataDomain, completeHash, simpleHash, responseFlag);

	 debug(string.format("Sending PROPAGATE_HASH to %s via %s", tostring(destination), tostring(channel)));
	 self:SendSections(channel, destination, section);
end

--[[ Sends an arbitray number of section strings to the given
--   destination via the given channel. If the channel is a public
--   one like, e.g. a guild channel, no destination has to be given.
--   A frame is only sent out if at least one section string is
--   given.
--   Returns: nil
--]]
function BuddyName:SendSections(channel, destination, ...)
	 argAssert(44, 1, IsNonEmptyString(channel), channel);
	 argAssert(44, 2, destination==nil or IsNonEmptyString(destination), destination);

	 -- If sync is not active, do nothing.
	 if (self:CfgGetSync()==false) then
			return;
	 end

	 local sections = { ... }; -- remaining function arguments

	 if (#sections == 0) then
			debug("SendSections(): Did not find any sections to send");
			return;
	 end

	 -- concatenate section strings
	 local sectionString = "";
	 for i = 1, #sections do
			sectionString = sectionString..sections[i];
	 end

	 -- compress complete section string
	 sectionString = self:Compress(sectionString, COMM.COMPRESSION.LIBCOMPRESS);
	 if (not sectionString) then
			Error(12, "SendSections: Could not compress string", 1);
			return;
	 end

	 -- build basic frame header
	 local headerString = string.format(COMM.PATTERNS.BUILD.FRAME_HEADER,
																			COMM.PROTOCOL,
																			#sections,
																			COMM.COMPRESSION.LIBCOMPRESS);

	 local frameString = headerString..sectionString;

	 BuddyName:SendCommMessage(COMM.SYNC_PREFIX, frameString, channel, destination);
end

-- TODO: document me
function BuddyName:BroadcastAllToGuild()
	 if (GetPlayerGuild()~=nil) then
			debug("BroadcastAllToGuild()");

			local shareEntries = {};
			local numEntries = 0;

			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p] and UnitIsInMyGuild(p)) then
						--debug(string.format("Sharing guild member %s",p));
						shareEntries[p] = v;
						numEntries=numEntries+1;
				 end
			end

			--commPrint(string.format(L["SYNC_SHARE_ENTRIES"], numEntries));

			if (numEntries==0) then
				 return
			end

			local entrySetString = CommEncodeEntrySet(shareEntries, COMM.ENTRY_SET_TYPE.FULL_DUMP);

			debug(string.format("Sending FULL_DUMP to GUILD"));
			self:SendSections("GUILD", nil, entrySetString);
	 end
end

--[[ Broadcasts a single buddy entry via the guild channel.
--   Won't do anything if player is not in a guild or
--   sync is deactivated.
--   Returns: nil
--]]
function BuddyName:BroadcastSingleEntryToGuild(entry)
	 argAssert(45, 1, IsTable(entry), entry);

	 if (self:CfgGetSync()==false or entry==nil) then
			return;
	 end

	 local sectionString = CommEncodeEntrySet(entry, COMM.ENTRY_SET_TYPE.SINGLE_ENTRY);

	 debug(string.format("Sending SINGLE_ENTRY to GUILD"));
	 self:SendSections("GUILD", nil, sectionString);
end

--[[ -- --------------------------------------------- -- ]]
--[[ -- Basic functions automatically called by Ace3  -- ]]
--[[ -- --------------------------------------------- -- ]]

--[[ Executed only once on startup
--   Returns: nil
--]]
function BuddyName:OnInitialize()
	 debug("OnInitialize called");

	 currentRealm = GetRealmName();

	 -- Register Database
	 self.db = LibStub("AceDB-3.0"):New("BuddyName_DB", cfgDefaults);

	 -- Register Options dialog
	 LibStub("AceConfig-3.0"):RegisterOptionsTable("BuddyNameOptions", self.optionsTable);
	 self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("BuddyNameOptions", "BuddyName"); -- second argument is displayed in options menu

	 -- Register Slash commands
	 for p,v in pairs(slashCommands) do
			if (v == true) then
				 self:RegisterChatCommand(p, "SlashProcessor")
			end
	 end

	 -- Create frame for events to hook up to and register battle net events
	 BN_EventFrame = CreateFrame("Frame", "BuddyNameFrame");
	 BN_EventFrame:SetScript("OnEvent", eventHandler);
	 for p,v in pairs(registerEvents) do
			if (v == true) then
				 BN_EventFrame:RegisterEvent(p);
			end
	 end

	 -- Register for receiving communication for synchronization purposes
	 BuddyName:RegisterComm(COMM.SYNC_PREFIX);
end


--[[ Executed each time the addon is switched on.
--   Returns: nil
--]]
function BuddyName:OnEnable()
	 debug("OnEnable() called");

	 -- if we're already active or config does not want us active, then quit
	 if (self:IsActive() or not self:CfgGetActive()) then
			return;
	 end

	 -- check if DB already exists, or if it is old and has to be converted
	 -- not done earlier because has to be called after PLAYER_ENTERING_WORLD (because of UpdateDB())
	 if (self.db.realm.version==nil) then -- if no version number, then either fresh db or very old. In each case, reset realm DB.
			self.db.realm.version=CONST.DB_VERSION -- set current version number
			self.db.realm.char = {}; -- set empty char table
	 else
			-- Apply fixes for broken DB entries
			if (self.db.realm.version <= 3) then
				 BuddyName:FixDB(1);
			end

			if (type(self.db.realm.cfg.guild[guildName])=="boolean") then
				 BuddyName:FixDB(2);
			end

			if (self.db.realm.version <= 3) then
				 BuddyName:FixDB(3);
			end

			if (self.db.realm.version < CONST.DB_VERSION) then -- update old database
				 BuddyName:UpdateDB();
			elseif (self.db.realm.version > CONST.DB_VERSION) then -- database is newer than BuddyName
				 statPrint(L["UPDATE_BUDDYNAME"]);
				 self:SetActive(false);
				 return;
			end
	 end

	 -- Hook into AddMessage
	 for i = 1, NUM_CHAT_WINDOWS do
			if (i~=2) then -- exclude combat log, better check against type than number
				 self:RawHook(_G["ChatFrame"..i], "AddMessage", "AddMessage", true);
			end
	 end

	 -- set our status to active
	 self:SetActive(true);

	 -- Initiate first guild sync 10 seconds after enabling BuddyName
	 -- Config status of guild sync is checked in called function
	 self:SetTimer("InitiateGuildSync", 10);
end


--[[ Executed each time the addon is switched off
--   Returns: nil
--]]
function BuddyName:OnDisable()
	 debug("OnDisable() called");

	 -- we're already inactive
	 if (not self:IsActive()) then
			return;
	 end

	 -- Unhook from AddMessage
	 self:Unhook("AddMessage");

	 -- set inactive
	 self:SetActive(false);
end


function BuddyName:Test()
end
